package java.nio;

import java.io.FileDescriptor;
import sun.misc.VM;
import sun.misc.Cleaner;
import sun.misc.Unsafe;
import sun.nio.ch.DirectBuffer;

class DirectByteBuffer extends MappedByteBuffer implements DirectBuffer
{
    protected static final Unsafe unsafe;
    private static final long arrayBaseOffset;
    protected static final boolean unaligned;
    private final Object att;
    private final Cleaner cleaner;
    
    @Override
    public Object attachment() {
        return this.att;
    }
    
    public Cleaner cleaner() {
        return this.cleaner;
    }
    
    DirectByteBuffer(int n) {
        super(-1, 0, n, n);
        int directMemoryPageAligned = VM.isDirectMemoryPageAligned();
        int pageSize = Bits.pageSize();
        long max = Math.max(1L, (long)n + (long)(directMemoryPageAligned ? pageSize : 0));
        Bits.reserveMemory(max, n);
        long allocateMemory;
        try {
            allocateMemory = DirectByteBuffer.unsafe.allocateMemory(max);
        }
        catch (OutOfMemoryError outOfMemoryError) {
            Bits.unreserveMemory(max, n);
            throw outOfMemoryError;
        }
        DirectByteBuffer.unsafe.setMemory(allocateMemory, max, (byte)0);
        if (directMemoryPageAligned && allocateMemory % (long)pageSize != 0L) {
            this.address = allocateMemory + (long)pageSize - (allocateMemory & (long)(pageSize - 1));
        }
        else {
            this.address = allocateMemory;
        }
        this.cleaner = Cleaner.create(this, new DirectByteBuffer.Deallocator(allocateMemory, max, n));
        this.att = null;
    }
    
    DirectByteBuffer(long address, int n, Object att) {
        super(-1, 0, n, n);
        this.address = address;
        this.cleaner = null;
        this.att = att;
    }
    
    private DirectByteBuffer(long address, int n) {
        super(-1, 0, n, n);
        this.address = address;
        this.cleaner = null;
        this.att = null;
    }
    
    protected DirectByteBuffer(int n, long address, FileDescriptor fileDescriptor, Runnable runnable) {
        super(-1, 0, n, n, fileDescriptor);
        this.address = address;
        this.cleaner = Cleaner.create(this, runnable);
        this.att = null;
    }
    
    DirectByteBuffer(DirectBuffer att, int n, int n2, int n3, int n4, int n5) {
        super(n, n2, n3, n4);
        this.address = att.address() + (long)n5;
        this.cleaner = null;
        this.att = att;
    }
    
    @Override
    public ByteBuffer slice() {
        int position = this.position();
        int limit = this.limit();
        int n = (position <= limit) ? (limit - position) : 0;
        int n2 = position << 0;
        assert n2 >= 0;
        return new DirectByteBuffer(this, -1, 0, n, n, n2);
    }
    
    @Override
    public ByteBuffer duplicate() {
        return new DirectByteBuffer(this, this.markValue(), this.position(), this.limit(), this.capacity(), 0);
    }
    
    @Override
    public ByteBuffer asReadOnlyBuffer() {
        return new DirectByteBufferR(this, this.markValue(), this.position(), this.limit(), this.capacity(), 0);
    }
    
    @Override
    public long address() {
        return this.address;
    }
    
    private long ix(int n) {
        return this.address + ((long)n << 0);
    }
    
    @Override
    public byte get() {
        return DirectByteBuffer.unsafe.getByte(this.ix(this.nextGetIndex()));
    }
    
    @Override
    public byte get(int n) {
        return DirectByteBuffer.unsafe.getByte(this.ix(this.checkIndex(n)));
    }
    
    @Override
    public ByteBuffer get(byte[] array, int n, int n2) {
        if ((long)n2 << 0 > 6L) {
            Buffer.checkBounds(n, n2, array.length);
            int position = this.position();
            int limit = this.limit();
            assert position <= limit;
            if (n2 > ((position <= limit) ? (limit - position) : 0)) {
                throw new BufferUnderflowException();
            }
            Bits.copyToArray(this.ix(position), array, DirectByteBuffer.arrayBaseOffset, (long)n << 0, (long)n2 << 0);
            this.position(position + n2);
        }
        else {
            super.get(array, n, n2);
        }
        return this;
    }
    
    @Override
    public ByteBuffer put(byte x) {
        DirectByteBuffer.unsafe.putByte(this.ix(this.nextPutIndex()), x);
        return this;
    }
    
    @Override
    public ByteBuffer put(int n, byte x) {
        DirectByteBuffer.unsafe.putByte(this.ix(this.checkIndex(n)), x);
        return this;
    }
    
    @Override
    public ByteBuffer put(ByteBuffer byteBuffer) {
        if ((byteBuffer instanceof DirectByteBuffer)) {
            if (byteBuffer == this) {
                throw new IllegalArgumentException();
            }
            DirectByteBuffer directByteBuffer = (DirectByteBuffer)byteBuffer;
            int position2 = directByteBuffer.position();
            int limit2 = directByteBuffer.limit();
            assert position2 <= limit2;
            int n2 = (position2 <= limit2) ? (limit2 - position2) : 0;
            int position3 = this.position();
            int limit3 = this.limit();
            assert position3 <= limit3;
            if (n2 > ((position3 <= limit3) ? (limit3 - position3) : 0)) {
                throw new BufferOverflowException();
            }
            DirectByteBuffer.unsafe.copyMemory(directByteBuffer.ix(position2), this.ix(position3), (long)n2 << 0);
            directByteBuffer.position(position2 + n2);
            this.position(position3 + n2);
        }
        else if (byteBuffer.hb != null) {
            int position = byteBuffer.position();
            int limit = byteBuffer.limit();
            assert position <= limit;
            int n = (position <= limit) ? (limit - position) : 0;
            this.put(byteBuffer.hb, byteBuffer.offset + position, n);
            byteBuffer.position(position + n);
        }
        else {
            super.put(byteBuffer);
        }
        return this;
    }
    
    @Override
    public ByteBuffer put(byte[] array, int n, int n2) {
        if ((long)n2 << 0 > 6L) {
            Buffer.checkBounds(n, n2, array.length);
            int position = this.position();
            int limit = this.limit();
            assert position <= limit;
            if (n2 > ((position <= limit) ? (limit - position) : 0)) {
                throw new BufferOverflowException();
            }
            Bits.copyFromArray(array, DirectByteBuffer.arrayBaseOffset, (long)n << 0, this.ix(position), (long)n2 << 0);
            this.position(position + n2);
        }
        else {
            super.put(array, n, n2);
        }
        return this;
    }
    
    @Override
    public ByteBuffer compact() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = (position <= limit) ? (limit - position) : 0;
        DirectByteBuffer.unsafe.copyMemory(this.ix(position), this.ix(0), (long)n << 0);
        this.position(n);
        this.limit(this.capacity());
        this.discardMark();
        return this;
    }
    
    @Override
    public boolean isDirect() {
        return true;
    }
    
    @Override
    public boolean isReadOnly() {
        return false;
    }
    
    @Override
    byte _get(int n) {
        return DirectByteBuffer.unsafe.getByte(this.address + (long)n);
    }
    
    @Override
    void _put(int n, byte x) {
        DirectByteBuffer.unsafe.putByte(this.address + (long)n, x);
    }
    
    private char getChar(long address) {
        if (DirectByteBuffer.unaligned) {
            int char1 = DirectByteBuffer.unsafe.getChar(address);
            return this.nativeByteOrder ? char1 : Bits.swap((char)char1);
        }
        return Bits.getChar(address, this.bigEndian);
    }
    
    @Override
    public char getChar() {
        return this.getChar(this.ix(this.nextGetIndex(2)));
    }
    
    @Override
    public char getChar(int n) {
        return this.getChar(this.ix(this.checkIndex(n, 2)));
    }
    
    private ByteBuffer putChar(long address, char c) {
        if (DirectByteBuffer.unaligned) {
            DirectByteBuffer.unsafe.putChar(address, this.nativeByteOrder ? c : Bits.swap(c));
            return this;
        }
        Bits.putChar(address, c, this.bigEndian);
        return this;
    }
    
    @Override
    public ByteBuffer putChar(char c) {
        this.putChar(this.ix(this.nextPutIndex(2)), c);
        return this;
    }
    
    @Override
    public ByteBuffer putChar(int n, char c) {
        this.putChar(this.ix(this.checkIndex(n, 2)), c);
        return this;
    }
    
    @Override
    public CharBuffer asCharBuffer() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = ((position <= limit) ? (limit - position) : 0) >> 1;
        if (!DirectByteBuffer.unaligned && (this.address + (long)position) % 2L != 0L) {
            return this.bigEndian ? new ByteBufferAsCharBufferB(this, -1, 0, n, n, position) : new ByteBufferAsCharBufferL(this, -1, 0, n, n, position);
        }
        return this.nativeByteOrder ? new DirectCharBufferU(this, -1, 0, n, n, position) : new DirectCharBufferS(this, -1, 0, n, n, position);
    }
    
    private short getShort(long address) {
        if (DirectByteBuffer.unaligned) {
            int short1 = DirectByteBuffer.unsafe.getShort(address);
            return this.nativeByteOrder ? short1 : Bits.swap((short)short1);
        }
        return Bits.getShort(address, this.bigEndian);
    }
    
    @Override
    public short getShort() {
        return this.getShort(this.ix(this.nextGetIndex(2)));
    }
    
    @Override
    public short getShort(int n) {
        return this.getShort(this.ix(this.checkIndex(n, 2)));
    }
    
    private ByteBuffer putShort(long address, short n) {
        if (DirectByteBuffer.unaligned) {
            DirectByteBuffer.unsafe.putShort(address, this.nativeByteOrder ? n : Bits.swap(n));
            return this;
        }
        Bits.putShort(address, n, this.bigEndian);
        return this;
    }
    
    @Override
    public ByteBuffer putShort(short n) {
        this.putShort(this.ix(this.nextPutIndex(2)), n);
        return this;
    }
    
    @Override
    public ByteBuffer putShort(int n, short n2) {
        this.putShort(this.ix(this.checkIndex(n, 2)), n2);
        return this;
    }
    
    @Override
    public ShortBuffer asShortBuffer() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = ((position <= limit) ? (limit - position) : 0) >> 1;
        if (!DirectByteBuffer.unaligned && (this.address + (long)position) % 2L != 0L) {
            return this.bigEndian ? new ByteBufferAsShortBufferB(this, -1, 0, n, n, position) : new ByteBufferAsShortBufferL(this, -1, 0, n, n, position);
        }
        return this.nativeByteOrder ? new DirectShortBufferU(this, -1, 0, n, n, position) : new DirectShortBufferS(this, -1, 0, n, n, position);
    }
    
    private int getInt(long address) {
        if (DirectByteBuffer.unaligned) {
            int int1 = DirectByteBuffer.unsafe.getInt(address);
            return this.nativeByteOrder ? int1 : Bits.swap(int1);
        }
        return Bits.getInt(address, this.bigEndian);
    }
    
    @Override
    public int getInt() {
        return this.getInt(this.ix(this.nextGetIndex(4)));
    }
    
    @Override
    public int getInt(int n) {
        return this.getInt(this.ix(this.checkIndex(n, 4)));
    }
    
    private ByteBuffer putInt(long address, int n) {
        if (DirectByteBuffer.unaligned) {
            DirectByteBuffer.unsafe.putInt(address, this.nativeByteOrder ? n : Bits.swap(n));
            return this;
        }
        Bits.putInt(address, n, this.bigEndian);
        return this;
    }
    
    @Override
    public ByteBuffer putInt(int n) {
        this.putInt(this.ix(this.nextPutIndex(4)), n);
        return this;
    }
    
    @Override
    public ByteBuffer putInt(int n, int n2) {
        this.putInt(this.ix(this.checkIndex(n, 4)), n2);
        return this;
    }
    
    @Override
    public IntBuffer asIntBuffer() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = ((position <= limit) ? (limit - position) : 0) >> 2;
        if (!DirectByteBuffer.unaligned && (this.address + (long)position) % 4L != 0L) {
            return this.bigEndian ? new ByteBufferAsIntBufferB(this, -1, 0, n, n, position) : new ByteBufferAsIntBufferL(this, -1, 0, n, n, position);
        }
        return this.nativeByteOrder ? new DirectIntBufferU(this, -1, 0, n, n, position) : new DirectIntBufferS(this, -1, 0, n, n, position);
    }
    
    private long getLong(long address) {
        if (DirectByteBuffer.unaligned) {
            long long1 = DirectByteBuffer.unsafe.getLong(address);
            return this.nativeByteOrder ? long1 : Bits.swap(long1);
        }
        return Bits.getLong(address, this.bigEndian);
    }
    
    @Override
    public long getLong() {
        return this.getLong(this.ix(this.nextGetIndex(8)));
    }
    
    @Override
    public long getLong(int n) {
        return this.getLong(this.ix(this.checkIndex(n, 8)));
    }
    
    private ByteBuffer putLong(long address, long n) {
        if (DirectByteBuffer.unaligned) {
            DirectByteBuffer.unsafe.putLong(address, this.nativeByteOrder ? n : Bits.swap(n));
            return this;
        }
        Bits.putLong(address, n, this.bigEndian);
        return this;
    }
    
    @Override
    public ByteBuffer putLong(long n) {
        this.putLong(this.ix(this.nextPutIndex(8)), n);
        return this;
    }
    
    @Override
    public ByteBuffer putLong(int n, long n2) {
        this.putLong(this.ix(this.checkIndex(n, 8)), n2);
        return this;
    }
    
    @Override
    public LongBuffer asLongBuffer() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = ((position <= limit) ? (limit - position) : 0) >> 3;
        if (!DirectByteBuffer.unaligned && (this.address + (long)position) % 8L != 0L) {
            return this.bigEndian ? new ByteBufferAsLongBufferB(this, -1, 0, n, n, position) : new ByteBufferAsLongBufferL(this, -1, 0, n, n, position);
        }
        return this.nativeByteOrder ? new DirectLongBufferU(this, -1, 0, n, n, position) : new DirectLongBufferS(this, -1, 0, n, n, position);
    }
    
    private float getFloat(long address) {
        if (DirectByteBuffer.unaligned) {
            int int1 = DirectByteBuffer.unsafe.getInt(address);
            return Float.intBitsToFloat(this.nativeByteOrder ? int1 : Bits.swap(int1));
        }
        return Bits.getFloat(address, this.bigEndian);
    }
    
    @Override
    public float getFloat() {
        return this.getFloat(this.ix(this.nextGetIndex(4)));
    }
    
    @Override
    public float getFloat(int n) {
        return this.getFloat(this.ix(this.checkIndex(n, 4)));
    }
    
    private ByteBuffer putFloat(long address, float n) {
        if (DirectByteBuffer.unaligned) {
            int floatToRawIntBits = Float.floatToRawIntBits(n);
            DirectByteBuffer.unsafe.putInt(address, this.nativeByteOrder ? floatToRawIntBits : Bits.swap(floatToRawIntBits));
            return this;
        }
        Bits.putFloat(address, n, this.bigEndian);
        return this;
    }
    
    @Override
    public ByteBuffer putFloat(float n) {
        this.putFloat(this.ix(this.nextPutIndex(4)), n);
        return this;
    }
    
    @Override
    public ByteBuffer putFloat(int n, float n2) {
        this.putFloat(this.ix(this.checkIndex(n, 4)), n2);
        return this;
    }
    
    @Override
    public FloatBuffer asFloatBuffer() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = ((position <= limit) ? (limit - position) : 0) >> 2;
        if (!DirectByteBuffer.unaligned && (this.address + (long)position) % 4L != 0L) {
            return this.bigEndian ? new ByteBufferAsFloatBufferB(this, -1, 0, n, n, position) : new ByteBufferAsFloatBufferL(this, -1, 0, n, n, position);
        }
        return this.nativeByteOrder ? new DirectFloatBufferU(this, -1, 0, n, n, position) : new DirectFloatBufferS(this, -1, 0, n, n, position);
    }
    
    private double getDouble(long address) {
        if (DirectByteBuffer.unaligned) {
            long long1 = DirectByteBuffer.unsafe.getLong(address);
            return Double.longBitsToDouble(this.nativeByteOrder ? long1 : Bits.swap(long1));
        }
        return Bits.getDouble(address, this.bigEndian);
    }
    
    @Override
    public double getDouble() {
        return this.getDouble(this.ix(this.nextGetIndex(8)));
    }
    
    @Override
    public double getDouble(int n) {
        return this.getDouble(this.ix(this.checkIndex(n, 8)));
    }
    
    private ByteBuffer putDouble(long address, double n) {
        if (DirectByteBuffer.unaligned) {
            long doubleToRawLongBits = Double.doubleToRawLongBits(n);
            DirectByteBuffer.unsafe.putLong(address, this.nativeByteOrder ? doubleToRawLongBits : Bits.swap(doubleToRawLongBits));
            return this;
        }
        Bits.putDouble(address, n, this.bigEndian);
        return this;
    }
    
    @Override
    public ByteBuffer putDouble(double n) {
        this.putDouble(this.ix(this.nextPutIndex(8)), n);
        return this;
    }
    
    @Override
    public ByteBuffer putDouble(int n, double n2) {
        this.putDouble(this.ix(this.checkIndex(n, 8)), n2);
        return this;
    }
    
    @Override
    public DoubleBuffer asDoubleBuffer() {
        int position = this.position();
        int limit = this.limit();
        assert position <= limit;
        int n = ((position <= limit) ? (limit - position) : 0) >> 3;
        if (!DirectByteBuffer.unaligned && (this.address + (long)position) % 8L != 0L) {
            return this.bigEndian ? new ByteBufferAsDoubleBufferB(this, -1, 0, n, n, position) : new ByteBufferAsDoubleBufferL(this, -1, 0, n, n, position);
        }
        return this.nativeByteOrder ? new DirectDoubleBufferU(this, -1, 0, n, n, position) : new DirectDoubleBufferS(this, -1, 0, n, n, position);
    }
    
    static {
        unsafe = Bits.unsafe();
        arrayBaseOffset = (long)DirectByteBuffer.unsafe.arrayBaseOffset(byte[].class);
        unaligned = Bits.unaligned();
    }
    
    private static class Deallocator implements Runnable
    {
        private static Unsafe unsafe;
        private long address;
        private long size;
        private int capacity;
        
        private Deallocator(long address, long size, int capacity) {
            assert address != 0L;
            this.address = address;
            this.size = size;
            this.capacity = capacity;
            return;
        }
        
        @Override
        public void run() {
            if (this.address == 0L) {
                return;
            }
            DirectByteBuffer.Deallocator.unsafe.freeMemory(this.address);
            this.address = 0L;
            Bits.unreserveMemory(this.size, this.capacity);
            return;
        }
        
        static {
            DirectByteBuffer.Deallocator.unsafe = Unsafe.getUnsafe();
        }
    }
}
